#!/usr/bin/perl

# ansys_launcher.pl - A launcher for various ANSYS programs
# (c) 2011-2016 Billy Constantine <william.constantine@adelaide.edu.au>

# How to use this file:
# - put it somewhere accessible to all users running ANSYS products
# - make symbolic links to it in /usr/local/bin or /opt/bin (or wherever)
#   with names "ansys", "ansyswb", "autodyn", "cfx", "fluent", "icemcfd" and
#   "lsdyna" (or whichever subset of those programs your users use)
# - optionally, also make links named "ansys12", "ansys13", "ansys14", etc,
#   to launch specific versions (assuming they're installed); and if you
#   have (e.g) both v14.0 and v14.5 installed, you can also make "ansys140",
#   "ansys145", etc, to be even more specific (the numbers correspond to
#   the version folders in your ansys installation directory with the leading
#   'v' omitted)

# The default version to launch can be globally specified using a default
# version file, described below

# Users can set their own default version with the default version
# environment variable

# If no default version is specified by either means, then the highest
# installed version will be used

use File::Basename qw(basename);
use Config;

$debug = 0;
# Debug mode
# More feedback about what's going on, and nothing will actually get run

$HOME = $ENV{HOME};
# User's home directory

############################################
# ADMINISTRATOR EDITS MAY BE REQUIRED HERE #
############################################

$ansys_inc = '/ansys_inc';
# Location of ANSYS installation

$default_version_variable = 'PREFERRED_ANSYS_VERSION';
$default_version_file = "/usr/local/share/default_ansys_version";
# Optional variable and file containing the default version to run
# 
# The variable or file should contain the two or three digit version number
# for the preferred default version, corresponding to the version-specific
# installation folder under $ansys_inc but without the leading 'v' (as
# above, in the instructions for creating symbolic links)
# 
# As a special case, if the variable or file is the string "LATEST", the
# latest available version will be used
# 
# Example 1: if you have v13.0, v14.0 and v14.5 installed, and you want
# v13.0 to be the default, then the default version file should contain
# "13" or "130"
# 
# Example 2: if you have the same three versions installed, and you want
# v14.5 to be the default, then the file should contain "14" or "145"
# 
# Example 3: if you have the same three versions installed, and you want
# v14.0 to be the default, then the file should contain "140"; this is
# because, if a two-digit version number is found, this launcher looks for
# the highest sub-version, so "14" would make v14.5 the default

$ENV{GC_MARKERS} ||= '';
# Often when starting an ANSYS application, this warning message appears:
# GC Warning: Limiting number of mark threads
# It is apparently a harmless enough message, but it has caused enough
# concern amongst my users that I try and make it go away. It seems to be
# enough just to have this variable defined in the environment, without
# even assigning it a value, to make that warning go away; so we assign
# it an empty value, unless it already has one

$bash    = '/bin/bash';
$uname26 = '/usr/bin/uname26';
$setarch = '/usr/bin/setarch';
# Location of the bash, uname, uname26 and setarch executables
# 
# The bash executable is used to run the ANSYS Workbench, due to issues
# with it being invoked via /bin/sh on some systems
# 
# uname26 and/or setarch are necessary for running ANSYS Workbench v14.0 and
# earlier on GNU/Linux systems with version 3 kernels; note that other ANSYS
# commands appear to be unaffected
# 
# You need either uname26 or setarch; you don't need both
# 
# uname26 can be obtained from Linux kernel mirrors in the people/ak/uname26
# directory (e.g. http://mirror.linux.org.au/linux/kernel/people/ak/uname26/)
# 
# setarch may or may not be distributed as standard by your GNU/Linux
# distribution

##############################
# END OF ADMINISTRATOR EDITS #
##############################

# $settings - the master configuration
# Each entry in $settings (with the exception of 'licensing') represents an
# ANSYS program
# An explanation of each program's sub-entries:
# - folder (path to where the program executable can be found; instances of
#           _VERSION_ in the path will be substituted with the ANSYS version
#           number; if a relative path is used, it will be considered relative
#           to $ansys_inc/v_VERSION_; if no path is specified,
#           $ansys_inc/_VERSION_/ansys/bin will be assumed)
# - program (the name of the exectuable file to launch; _VERSION_ will be
#            substituted with the ANSYS version number; this setting is
#            optional (if it is not set, then the name used to launch this
#            program will be used instead))
# - arguments (command line arguments for the executable that should always
#              be used; include these as a single string, e.g. '-a -b -c=d';
#              user-specified arguments will be added to the end of these
#              arguments when the program is launched; this setting is
#              optional)
# - environment (environment variables that should be set; use a comma-
#                separated list of '<VARIABLE>=<value>'-style entries, e.g.
#                'DEBUG=0,LANG=en_US'; you can also say '<VARIABLE>^=<value>'
#                or '<VARIABLE>.=<value>' to respectively prepend or append
#                the value to an existing variable; _FOLDER_ will be
#                substituted with the application's 'folder' setting; this
#                setting is optional)
# 
# The 'licensing' entry is special; it isn't a program entry, it's just
# there so we can add the folder to PATH so that the license utilities are
# available within the program (if required)

$settings = {
             'licensing'    => {
                                'folder'      => "$ansys_inc/shared_files/licensing/lic_admin",
                                },
             'ansys'        => {
                                'program'     => 'launcher_VERSION_',
                                },
             'ansyswb'      => {
                                'folder'      => 'Framework/bin/Linux64',
                                'program'     => 'runwb2',
                                'arguments'   => '-oglmesa',
                                'environment' => 'MALLOC_CHECK_=0,LC_NUMERIC=C,LANG=en_US.UTF-8,HWLOC_HIDE_ERRORS=1',
                              },
             'autodyn'      => {
                                'folder'      => 'autodyn/bin',
                                'program'     => 'autodyn_VERSION_',
                                },
             'chemkin'      => {
                                'folder'      => 'reaction/chemkinpro.linuxx8664/bin',
                                'program'     => 'run_chemkin.sh',
                                'arguments'   => 'Pro',
                                },
             'chemkin_post' => {
                                'folder'      => 'reaction/chemkinpro.linuxx8664/bin',
                                'program'     => 'run_chemkin.sh',
                                },
             'cfx'          => {
                                'folder'      => 'CFX/bin',
                                'program'     => 'cfx5',
                                },
             'energico'     => {
                                'folder'      => 'reaction/chemkinpro.linuxx8664/bin',
                                'program'     => 'run_energico.sh',
                                },
             'fluent'       => {
                                'folder'      => 'fluent/bin',
                                'arguments'   => '-ssh',
                                'environment' => 'HWLOC_HIDE_ERRORS=1',
                                },
             'forte'        => {
                                'folder'      => 'reaction/forte.linuxx8664/bin',
                                'program'     => 'forte.sh',
                                },
             'icemcfd'      => {
                                'folder'      => 'icemcfd/linux64_amd/bin',
                                },
             'lsdyna'       => {
                                'program'     => 'lsdyna_VERSION_',
                                },
             'rdworkbench'  => {
                                'folder'      => 'reaction/chemkinpro.linuxx8664/bin',
                                'program'     => 'run_rdworkbench.sh',
                                'arguments'   => 'Pro',
                                },
             };

$exe = basename($0);
($app, $version) = $exe =~ /^(.+\D)(\d{2,3})?$/;

for(keys(%{$settings})) {
   if(0 == length($settings->{$_}->{folder})) {
      if($app eq $_) {
         warn("$0: no folder defined for application \"$_\"; defaulting to ansys/bin\n");
      }
      $settings->{$_}->{folder} = 'ansys/bin';
   }
   # if no folder is defined for the application, set it to the default
   if(!($settings->{$_}->{folder} =~ m@^/@)) {
      $settings->{$_}->{folder} =~ s@^@$ansys_inc/v_VERSION_/@;
   }
   # convert relative paths to full paths
}

if(!(exists($settings->{$app}))) {
   # if run with an unrecognised name, check command line arguments; if the
   # first one is '-applist', then list the known applications and the full
   # paths to their executables and quit; this is to help the install script
   if($ARGV[0] eq '-applist') {
      print("ansys_inc,$ansys_inc\n");
      for(sort(keys(%{$settings}))) {
         next if $_ eq 'licensing';
         print("$_,",
               join('/', ($settings->{$_}->{folder},
                          (exists($settings->{$_}->{program}) ?
                            $settings->{$_}->{program} : $_))), "\n");
      }
      exit(0);
   } else {
      die("$0: couldn't determine an application to launch from \"$exe\"\n",
          "Known applications are:  ",
          join('  ', sort(grep { $_ ne 'licensing' }
                            keys(%{$settings}))), "\n");
   }
} elsif(!(exists($settings->{$app}->{folder}))) {
   die("$0: no folder specified for application \'$app\"\n");
}

# Get a list of all the available versions of ANSYS
opendir(ANSYS_INC, $ansys_inc)
  || die("$0: couldn't opendir() \"$ansys_inc\": $!");
$debug && warn("$0: searching \"$ansys_inc\" for installed versions\n");
while($_ = readdir(ANSYS_INC)) {
   next unless -d "$ansys_inc/$_";
   next unless /^v(\d\d)(\d+)$/;
   $major = $1;
   $full  = $1 . $2;
   $debug && warn("$0: found $full (major version $major)\n");
   ++$installed_versions->{FULL}->{$full};
   ++$installed_versions->{MAJOR}->{$major}->{FULL}->{$full};
   if($installed_versions->{LATEST}->{FULL} < $full) {
      $debug
        && warn("$0: replacing latest full version $installed_versions->{LATEST}->{FULL} with $full\n");
      $installed_versions->{LATEST}->{FULL} = $full;
   }
   if($installed_versions->{LATEST}->{MAJOR} < $major) {
      $debug
        && warn("$0: replacing latest major version $installed_versions->{LATEST}->{MAJOR} with $major\n");
      $installed_versions->{LATEST}->{MAJOR} = $major;
   }
   if($installed_versions->{MAJOR}->{$major}->{LATEST} < $full) {
      $debug
        && warn("$0: replacing latest full version $installed_versions->{MAJOR}->{$major}->{LATEST} for $major with $full\n");
      $installed_versions->{MAJOR}->{$major}->{LATEST} = $full;
   }
}
closedir(ANSYS_INC);

if(0 == scalar(keys(%{$installed_versions->{FULL}}))) {
   die("$0: fatal error: no versions of ANSYS installed in \"$ansys_inc\"");
}

# Now we can determine:
# - the latest major version
#      $installed_versions->{LATEST}->{MAJOR}
# - the latest full version of major version XX
#      $installed_versions->{MAJOR}->{XX}->{LATEST}
# - the latest full version
#      $installed_versions->{LATEST}->{FULL}
#      $installed_versions->{MAJOR}->{$installed_versions->{LATEST}->{MAJOR}}->{LATEST}
# - the major versions that are available
#      keys(%{$installed_versions->{MAJOR}})
# - the full versions that are available
#      keys(%{$installed_versions->{FULL}})
# - the full versions that are available for major version XX
#      keys(%{$installed_versions->{MAJOR}->{XX}->{FULL}})
# - if major version XX is installed
#      exists($installed_versions->{MAJOR}->{XX})
# - if full version XXX is installed
#      exists($installed_versions->{FULL}->{XXX})

if(exists($ENV{$default_version_variable})) {
   $debug && warn("$0: getting default ansys version from env\n");
   $default_version = $ENV{$default_version_variable};
} elsif(-f $default_version_file && -r $default_version_file) {
   $debug && warn("$0: getting default ansys version from file\n");
   if(open(DVF, $default_version_file)) {
      while(<DVF>) {
         s/\s//g;
         next unless /\S/;
         $default_version = $_;
         last;
      }
      close(DVF);
   } else {
      warn("$0: couldn't open \"$default_version_file\": $!\n",
           "$0: will use latest installed version\n");
      $default_version = 'LATEST';
   }
} else {
   $debug &&
     warn("$0: getting default ansys version by searching $ansys_inc\n");
   $default_version = 'LATEST';
}

$debug && warn("$0: default version is \"$default_version\"\n");
if($default_version eq 'LATEST') {
   $default_version = $installed_versions->{LATEST}->{FULL};
   $debug && warn("$0: default version set to \"$default_version\"\n");
}

$version =~ s/^v//i;

if($version
  && (!(exists($installed_versions->{MAJOR}->{$version})
     || exists($installed_versions->{FULL}->{$version})))) {
   warn("$0: requested version \"$version\" is not available\n");
   warn("$0: using default version \"$default_version\" instead\n");
   $version = $default_version;
}

if(length($version) < 3) {
   if(0 == length($version)) {
      warn("$0: no version number specified: using default \"$default_version\"\n");
      $version = $default_version;
   }

   if(2 == length($version)) {
      warn("$0: short version number specified: using latest sub-version \"$installed_versions->{MAJOR}->{$version}->{LATEST}\"\n");
      $version = $installed_versions->{MAJOR}->{$version}->{LATEST};
   }
   warn("$0: use \"$app$version\" to run this specific version\n");
}

if(!(exists($installed_versions->{FULL}->{$version}))) {
   die("$0: requested version \"$version\" is not available\n$0: no recovery is possible, contact your system administrator\n");
}

$settings->{$app}->{folder}  =~ s@_VERSION_@$version@g;
$settings->{$app}->{program} =~ s@_VERSION_@$version@g;

if(0 == length($settings->{$app}->{program})) {
   @possible_progs = ($exe);
   if($app ne $exe) {
      unshift(@possible_progs, $app);
   }
   warn("$0: no program specified, will check for ",
        join(' and ', @possible_progs), "\n");
   for $p (@possible_progs) {
      $fullp = $settings->{$app}->{folder} . '/' . $p;
      if(!(-e $fullp)) {
         warn("$0: \"$p\": file does not exist\n");
      } elsif(!(-f $fullp)) {
         warn("$0: \"$p\": file is not a regular file\n");
      } elsif(!(-x $fullp)) {
         warn("$0: \"$p\": file is not executable\n");
      } else {
         warn("$0: using \"$p\"\n");
         $settings->{$app}->{program} = $p;
         last;
      }
   } 
}

if(0 == length($settings->{$app}->{program})) {
   die("$0: unable to determine an executable file to use\n");
}

undef($error);
if(!(-d $settings->{$app}->{folder})) {
   $error = "$0: \"$settings->{$app}->{folder}\": folder does not exist\n";
} elsif(!(-e $settings->{$app}->{folder}.'/'.$settings->{$app}->{program})) {
   $error = "$0: \"$settings->{$app}->{program}\": file does not exist\n";
} elsif(!(-f $settings->{$app}->{folder}.'/'.$settings->{$app}->{program})) {
   $error = "$0: \"$settings->{$app}->{program}\": file is not a regular file\n";
} elsif(!(-x $settings->{$app}->{folder}.'/'.$settings->{$app}->{program})) {
   $error = "$0: \"$settings->{$app}->{program}\": file is not executable\n";
}
if($error) {
   warn($error);
   if(!($debug)) {
      exit(1);
   }
}

$ENV{PATH} = join(':',
                  ($settings->{$app}->{folder},
                   $settings->{licensing}->{folder},
                   $ENV{PATH}));

$ld_library_path_adjustments = $settings->{$app}->{folder};
for $l (qw(Mesa)) {
   if(-d "$settings->{$app}->{folder}/$l") {
      $ld_library_path_adjustments .= ":$settings->{$app}->{folder}/$l";
   }
}
$ENV{LD_LIBRARY_PATH} = join(':',
                             ($ld_library_path_adjustments,
                              $ENV{LD_LIBRARY_PATH}));

if(length($settings->{$app}->{environment})) {
   $settings->{$app}->{environment} =~
     s@_FOLDER_@$settings->{$app}->{folder}@g;
   @F = split(/\,/, $settings->{$app}->{environment});
   for(@F) {
      ($var, $op, $val) = /^([^=]+?)([\.\^]?=)(.+)$/;
      if((!(exists($ENV{$var})) || ($op eq '='))) {
         $ENV{$var} = $val;
         if($debug) {
            warn("setting \"$var\" to \"$val\"\n");
            if($op ne '=') {
               warn("(\"$var\" had no value previously)\n")
            }
         }
      } elsif($op eq '^=') {
         $ENV{$var} =~ s@^@$val:@;
         if($debug) {
            warn("prepending \"$val\" to \"$var\"\n");
         }
      } elsif($op eq '.=') {
         $ENV{$var} =~ s@$@:$val@;
         if($debug) {
            warn("appending \"$val\" to \"$var\"\n");
         }
      }
   }
}

if(length($settings->{$app}->{arguments})) {
   unshift(@ARGV, split(/\s+/, $settings->{$app}->{arguments}));
}
unshift(@ARGV, $settings->{$app}->{folder}.'/'.$settings->{$app}->{program});

if($app eq 'ansyswb' || $app eq 'fluent') {
   $ENV{FLUENT_ARCH} ||= 'lnamd64';
}

if($app eq 'ansyswb') {
   if(0 == length($bash)) {
      die("$0: fatal error: \$bash undefined\n");
   } elsif(!(-f $bash)) {
      die("$0: fatal error: cannot find \'$bash\'\n");
   } elsif(!(-x $bash)) {
      die("$0: fatal error: \'$bash\' is not executable\n");
   }
   unshift(@ARGV, ($bash, '--norc'));
}

if($version <= 140 && $app eq 'ansyswb' &&
     $Config{osname} eq 'linux' && $Config{osvers} >= 3) {
   warn("$0: will need to force kernel 2.6 personality\n");
   $k26p = undef;
   if((0 == length($uname26)) && (0 == length($setarch))) {
      die("$0: fatal error: \$uname26 and \$setarch undefined\n");
   } else {
      for $k ($setarch, $uname26) {
         if(0 != length($k) && -f $k && -x $k) {
            $k26p = $k;
            last;
         }
      }
   }
   if(0 == length($k26p)) {
      die("$0: fatal error: cannot find means of forcing kernel 2.6 personality\n");
   }
   if($k26p eq $setarch) {
      unshift(@ARGV, qw(x86_64 --uname-2.6));
   }
   unshift(@ARGV, $k26p);
}

if($debug) {
   warn("$0: would launch:\n\t\"",
        join('", "', @ARGV), "\"\n");
} else {
   exec(@ARGV);
}
